#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import logging

from factdatasource.execptions import NotSupportedFDException, FDExecuteException
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext
from libs.utils import get_algo_abs_path, delete_file
from munch import Munch
from api.model.AlgoOutWrapper import AlgoOutWrapper
import os

log = logging.getLogger('Ficus')


class AlgorithmDatasourceContext(AbstractFactDatasourceContext):

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        fd = self.fd()
        file_name = get_algo_abs_path(fd.site, fd.connection, fd.target)
        if not os.path.exists(file_name):
            return 0
        elif os.path.isfile(file_name):
            return 1
        else:
            # 只计算一层
            return len(os.listdir(file_name))

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        return self.size() == 0

    def collect(self, size: int):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """
        raise NotSupportedFDException('algorithm不支持该操作')

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        raise NotSupportedFDException('algorithm不支持该操作')

    def query(self, query: str, parameters: dict = None):
        """
        使用查询语句查询数据
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        raise NotSupportedFDException('algorithm不支持该操作')

    def inserts(self, result_list: list):
        """
        insert是重新写入
        :param result_list: 要保存的数据
        :return:
        """
        return self.__save_or_update(result_list, 'insert')

    def updates(self, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        return self.__save_or_update(result_list, 'update')

    def inserts_or_updates(self, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        return self.__save_or_update(result_list, 'saveOrUpdate')

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        # 消息可能是有顺序的,因此先不并发
        raise NotSupportedFDException('algorithm不支持该操作')

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        # 消息可能是有顺序的,因此先不并发
        raise NotSupportedFDException('algorithm不支持该操作')

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        # 消息可能是有顺序的,因此先不并发
        raise NotSupportedFDException('algorithm不支持该操作')

    def delete_all(self):
        """
        清空数据
        :return:
        """
        fd = self.fd()
        delete_file(get_algo_abs_path(fd.site, fd.connection, ''))

    def delete(self, query: str):
        """
        query默认是个文件名称，删除这个fd配置路径下的该文件
        :return:
        """
        fd = self.fd()
        delete_file(get_algo_abs_path(fd.site, fd.connection, query))

    def delete_conditions(self, condition_groups: list):
        raise NotSupportedFDException('algorithm不支持该操作')

    def __save_or_update(self, result_list: list, method: str):
        if not result_list:
            return []
        fd = self.fd()
        results = []
        result_one = result_list[0]
        if isinstance(result_one, dict) and 'subPath' in result_one.keys():
            # 说明是AlgoOutWrapper.WITH传进来的
            for serializable in result_list:
                try:
                    if not isinstance(serializable, dict):
                        serializable = vars(serializable)
                    data = serializable.get('t')
                    sub_path = serializable.get('subPath')
                    abs_sub_path = get_algo_abs_path(fd.site, fd.connection, sub_path)
                    self.__validate_file(method, abs_sub_path)
                    with open(abs_sub_path, 'ab+') as f:
                        f.write(self.__getByteData(data))
                    results.append(Munch({"success": True}))
                except Exception as e:
                    error = 'AlgorithmFactDatasourceContext插入时，文件写入报错'
                    log.error(error, e)
                    results.append(Munch({"success": False, "error": f'AlgorithmFactDatasourceContext插入时，文件写入报错，路径是{abs_sub_path}',
                                          "content": serializable}))
        else:
            # 说明是普通模式
            if not fd.target:
                raise FDExecuteException('fd的target为空，不是文件类型，不允许插入')
            path = get_algo_abs_path(fd.site, fd.connection, fd.target)
            self.__validate_file(method, path)
            for serializable in result_list:
                try:
                    with open(path, 'ab+') as f:
                        f.write(self.__getByteData(serializable))
                    results.append(Munch({"success": True}))
                except Exception as e:
                    error = 'AlgorithmFactDatasourceContext插入时，文件写入报错'
                    log.error(error, e)
                    results.append(Munch({"success": False, "error": f'AlgorithmFactDatasourceContext插入时，文件写入报错，路径是{path}',
                                          "content": serializable}))
        return results

    def __validate_file(self, method: str, path: str):
        if not os.path.exists(path):
            if 'update' == method:
                raise FDExecuteException(f'{path}路径的文件不存在，无法更新')
            parent_dir = os.path.abspath(os.path.dirname(path))
            if not os.path.exists(parent_dir):
                os.makedirs(parent_dir)
            f = open(path, 'w')
            f.close()
        else:
            # 只有插入的是新建文件，需要删除旧文件
            if 'insert' == method:
                os.remove(path)
                f = open(path, 'w')
                f.close()

    def __getByteData(self, data):
        import json
        if isinstance(data, dict):
            return json.dumps(data).encode()
        elif isinstance(data, str):
            return data.encode()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
import json
import logging
import threading
import time

from confluent_kafka.cimpl import KafkaError, KafkaException
from munch import Munch

from api.model.FactDatasource import FactDatasource, FactDatasourceTypeEnum
from config.annotation import Value
from factdatasource import InnerFdCode, fold_ref_fds
from factdatasource.FactDatasourceContextHolder import FactDatasourceContextHolder
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from factdatasource.dao.kafka.MultipleKafkaDatasource import CustomKafkaClient
from libs.utils import Singleton

log = logging.getLogger('Ficus')


def wait_using_fd(fd_context_holder: FactDatasourceContextHolder, action: int, fact_datasource: FactDatasource):
    def listen_using_fd():
        fd_code = fact_datasource.code
        if fd_context_holder.is_using_fact_datasource(fd_code):
            while fd_context_holder.is_using_fact_datasource(fd_code):
                log.debug(f'监听到数据源{fd_code}的{action}事件，等待数据源使用完毕')
                time.sleep(0.5)

        if action == FactDatasourceChangeListener.ACTION_UPDATE:
            fd_context_holder.update_fact_datasource(fact_datasource)
        elif action == FactDatasourceChangeListener.ACTION_DELETE:
            fd_context_holder.remove_fact_datasource(fact_datasource)

    thread = threading.Thread(name='wait_using_fd', target=listen_using_fd, daemon=True)
    thread.start()


class FactDatasourceChangeListener(threading.Thread, Singleton):
    """
    监听数据源消息
    """
    ACTION_ADD = 0
    ACTION_UPDATE = 1
    ACTION_DELETE = 2

    FACT_DATASOURCE_CHANGE_TOPIC = 'springCloudBus'

    @Value("${actor.name:unknown}")
    def actor_name(self):
        pass

    def __init__(self):
        threading.Thread.__init__(self)

        client: CustomKafkaClient = get_multiple_datesource(FactDatasourceTypeEnum.KAFKA).get_client()
        self.producer = client.producer
        self.consumer = client.consumer
        self.running = False
        # 守护线程运行
        self.daemon = True

    @property
    def fd_context_holder(self):
        return FactDatasourceContextHolder.instance()

    def run(self):
        if self.running:
            log.warning(f'FD监听线程已启动无需重复启动')
            return
        self.running = True

        # 这里关心已使用过的FD的修改和删除消息，添加应该不需要关心
        # 这里监听到消息有两件事， 1. fd_context的事件处理  2. 多数据源的事件处理
        # 3.修改和删除需要判断该数据源是否正在使用，如果正在使用需要等待使用完成后在进行修改
        self.consumer.subscribe([self.FACT_DATASOURCE_CHANGE_TOPIC])

        while self.running:
            try:
                msg = self.consumer.poll(timeout=1.0)
                if msg is None:
                    continue
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        continue
                    else:
                        # Error
                        raise KafkaException(msg.error())
                else:
                    # 进行消息的处理
                    message = Munch(json.loads(msg.value()))
                    # 只处理FactDatasourceChangeEvent事件
                    if message.type == 'FactDatasourceChangeEvent':
                        self.deal_message(message)
            except Exception as e:
                log.exception(f"监听FactDatasourceChange消息出现问题,", e)
            finally:
                if self.running == False:
                    # 把消息队列关闭掉
                    self.consumer.close()

    def stop(self):
        self.running = False

    def deal_message(self, change_event):
        action = change_event.action

        # 需要判断是否是这个actor所关心的fd

        # TODO 这里有问题 event里面的fd的 type有可能是ref的.这就导致了后面卸载的时候卸载不了
        #  还有一个问题 那就是 A->ref B   然后某个CE是关联的A.  那么B变更了后,现在这里是不会处理的
        #  要解决这个问题,就需要和JAVA的一样, 要找到这个执行器所关心的FD,并且把ref的也要找出来.处理.
        #  也就是这个类要重写
        fact_datasource = FactDatasource(**change_event.factDatasource)

        if action == self.ACTION_ADD:
            # 不关心添加数据源事件
            return

        # 这里就需要处理了
        actor_name = self.actor_name()
        # 找到这个执行器所有关心的fds
        actor_fds = self.find_actor_fds(actor_name)

        # 这个就是这个执行器所关心的所有的FD
        from client import FactDatasourceManageClient
        fact_datasources = [
            FactDatasourceManageClient.get(inner_fd_code.site, inner_fd_code.project, inner_fd_code.code)
            for inner_fd_code in actor_fds]

        # 把用户关心的FD都查询一遍,找到ref
        actorFdsWithRef = set()
        for datasource in fact_datasources:
            if datasource.type == "REF" and datasource.ref is not None:
                tmp = list(map(lambda fd: InnerFdCode(fd.site, fd.projectCode, fd.code), fold_ref_fds(datasource)))
                actorFdsWithRef.update(tmp)
            else:
                tmp = [InnerFdCode(datasource.site, datasource.projectCode, datasource.code)]
                actorFdsWithRef.update(tmp)

        # 判断是否是这个actor所关心的
        any = list(filter(lambda
                              actor_fd: actor_fd.site == fact_datasource.site and actor_fd.project == fact_datasource.projectCode and actor_fd.code == fact_datasource.code,
                          actorFdsWithRef))
        # 如果这里的any有值,说明满足要求
        if any is None or len(any) == 0:
            # 说明这个执行器不关心这个FD的变化,不进行处理
            return

        # 这里需要找到这个fd变化后 影响到的 fds
        changed_fds = set()
        from client import FactDatasourceClient
        for datasource in fact_datasources:
            # 第二次做遍历
            if datasource.type == "REF" and datasource.ref is not None:
                tmp = fold_ref_fds(datasource)
                rrr = list(filter(lambda
                                      actor_fd: actor_fd.site == fact_datasource.site and actor_fd.projectCode == fact_datasource.projectCode and actor_fd.code == fact_datasource.code,
                                  tmp))
                if rrr is not None and len(rrr) > 0:
                    changed_fds.add(FactDatasource(**FactDatasourceClient.fd(datasource.code)))       # TODO 这里是有问题的,应该使用 FactDatasourceClient.fd(fd_code)
            else:
                if datasource.site == fact_datasource.site and datasource.projectCode == fact_datasource.projectCode and datasource.code == fact_datasource.code:
                    changed_fds.add(FactDatasource(**FactDatasourceClient.fd(datasource.code)))   # TODO 这里是有问题的,应该使用 FactDatasourceClient.fd(fd_code)

        for fd in changed_fds:
            if not self.fd_context_holder.is_loaded_fact_datasource(fd.code):
                # 本地数据源还没有使用不进行处理
                continue

            if self.fd_context_holder.is_using_fact_datasource(fd.code):
                wait_num = 0
                while self.fd_context_holder.is_using_fact_datasource(fd.code):
                    log.debug(f'监听到数据源{fd.code}的{action}事件，等待数据源使用完毕')
                    wait_num += 1
                    time.sleep(0.5)
                    # 5秒都还在使用的话就只能使用新线程来处理了,否则无法处理后续的事件
                    if wait_num > 10:
                        wait_using_fd(self.fd_context_holder, action, fd)
                        break

            log.info(f"FD({fd.get_source_name()}接收到变化:{action} 开始刷新)")

            if action == self.ACTION_UPDATE:
                self.fd_context_holder.update_fact_datasource(fd)
            elif action == self.ACTION_DELETE:
                self.fd_context_holder.remove_fact_datasource(fd)

    def find_actor_fds(self, actor_name) -> set:
        """
        找到这个actor所有的jobs,并找到他们的fdCode
        :param actor_name:
        :return:
        """

        # 找到这个actor所对应的所有的jobs
        from client import JobScheduleClient
        jobs = JobScheduleClient.get_jobs(actor_name)

        if jobs is None or len(jobs) == 0:
            return set()

        return self.__generate_fd_codes(jobs)

    def __generate_fd_codes(self, jobs: list) -> set:
        """
        构造出job的所有FD信息
        :param jobs:
        :return:
        """
        fdCodes = set()

        for job in jobs:
            actorParam = job.actorParam
            if actorParam is None or actorParam == "":
                continue
            try:
                json_object = json.loads(actorParam)
                code_ = json_object["code_"]
                site_ = json_object["site_"]
                projectCode_ = json_object["projectCode_"]
                from client import ComputeExecutionClient
                dataComputeExecution = ComputeExecutionClient.get(site_, projectCode_, code_)
                if dataComputeExecution is None:
                    # 有可能是crawl
                    from client import DataCrawlClient
                    dataCrawl = DataCrawlClient.get(site_, projectCode_, code_)
                    if dataCrawl is None:
                        # 既不是ce又不是crawl,跳过
                        continue
                    fdCodes.update(self.__innerAddFdCodes(site_, projectCode_, dataCrawl.outputFdCodes))
                else:
                    fdCodes.update(self.__innerAddFdCodes(site_, projectCode_, dataComputeExecution.sourceFdCodes))
                    fdCodes.update(self.__innerAddFdCodes(site_, projectCode_, dataComputeExecution.outputFdCodes))
            except:
                pass

        return fdCodes

    def __innerAddFdCodes(self, site_, projectCode_, outputFdCodes: list) -> list:
        if outputFdCodes is None or len(outputFdCodes) == 0:
            return list()

        return list(map(lambda code: InnerFdCode(site_, projectCode_, code), outputFdCodes))



"""
Algorithm的操作模块
"""
import requests
from munch import Munch

from api.exceptions import ServiceInnerException, AuthException
from client import check_instance_avaliable, do_service

def get(code, actor_name):
    """
    获取Algorithm
    :param code:
    :param actor_name:
    :return: 如果这个Algorithm存在,返回Algorithm对象.否则返回空
    """
    check_instance_avaliable()

    try:
        r = do_service(f"/v1/algorithm/search", params = {"code" : code, "actorName": actor_name}, return_type="json")

        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e

def exists(code, actor_name):
    """
    获取Algorithm
    :param code:
    :param actor_name:
    :return: 如果这个Algorithm存在,返回True.否则返回False
    """
    check_instance_avaliable()

    try:
        r = do_service(f"/v1/algorithm/exists", params = {"code" : code, "actorName": actor_name})

        if r is not None:
            return r
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e

def add(algoithm):
    """
    新增Algorithma
    :param algoithm对象
    :return: 如果添加失败,就抛出异常,否则无返回
    """
    if algoithm is None:
        return

    check_instance_avaliable()

    request = algoithm
    if isinstance(algoithm, Munch):
        # 说明是munch的,那么就转成Dict的
        request = algoithm.toDict()

    try:
        r = do_service(f"/v1/algorithm", method="post", data=request, return_type="none")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e

def update(algoithm):
    """
    更新Algorithma
    :param algoithm对象
    :return: 如果添加失败,就抛出异常,否则无返回
    """
    if algoithm is None:
        return

    check_instance_avaliable()

    request = algoithm
    if isinstance(algoithm, Munch):
        # 说明是munch的,那么就转成Dict的
        request = algoithm.toDict()

    try:
        r = do_service(f"/v1/algorithm/update", method="put", data=request, return_type="none")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
from munch import Munch


class InnerFdCode(object):

    def __init__(self, site: str, project: str, code: str):
        self.site = site
        self.project = project
        self.code = code

    def __hash__(self):
        result = 1
        result = 31 * result + (hash(self.site) if self.site is not None else 0)
        result = 31 * result + (hash(self.project) if self.project is not None else 0)
        result = 31 * result + (hash(self.code) if self.code is not None else 0)
        return result

    def source_name(self):
        return f"fd_{self.site}_{self.project}_{self.code}"

    def __eq__(self, other):
        return self.site == other.site and self.project == other.project and self.code == other.code


class InnerCrawl(InnerFdCode):

    def __init__(self, site: str, project: str, code: str, type: str, connection: str, credentials: str):
        super().__init__(site, project, code)
        self.connection = connection
        self.type = type
        self.credentials = credentials

    def source_name(self):
        return f"crawl_{self.site}_{self.project}_{self.code}"


def fold_ref_fds(factDatasource, result: list = None) -> list:
    """
    把Ref的FactDatasource平面的展开
    :param result:
    :param factDatasource:
    :return:
    """
    if result is None:
        result = list()

    result.append(factDatasource)
    if "ref" in factDatasource and len(factDatasource["ref"]) != 0:
        fold_ref_fds(Munch(factDatasource.ref), result)

    return result



import logging
import time
import uuid

from api.exceptions import IllegalArgumentException
from client import DataAlgorithmClient

log = logging.getLogger('Ficus')

def check_ficus_online() -> bool:
    """
    检查ficus服务是否存在
    :return:
    """
    from client import check_instance_avaliable
    from api.exceptions import ServiceNoInstanceException
    try:
        check_instance_avaliable()
        return True
    except ServiceNoInstanceException:
        return False

def build_param_def(conf_param_defs:list, is_input_param:bool) -> list:
    """
    构建出参入参
    :param conf_param_defs: 配置文件读取的出/入参
    :param is_input_param: 是否如此
    :return:
    """
    param_defs = list()
    for conf_param_def in conf_param_defs:
        # 获取配置文件读取的字段信息
        if "model" in conf_param_def and len(conf_param_def["model"]) > 0:
            param_def = {
                "type": conf_param_def.get("type").upper(),
                "description": conf_param_def.get("description"),
                "matchTags": conf_param_def.get("matchTags"),
                "matchParams": conf_param_def.get("matchParams"),
                "isInputParam": is_input_param,
                "createTime": time.strftime("%Y-%m-%d %H:%M:%S"),
                "metaModel": {
                    "site": "S1",
                    "project_code": "__NO_PROJECT__",
                    "code": str(uuid.uuid4()).replace('-', ''),
                    "description":"算法组件模型创建",
                    "fields": list()
                }
            }

            for conf_field in conf_param_def["model"]:
                meta_model_field = {
                    "fieldName": conf_field.get("fieldName"),
                    "description": conf_field.get("description"),
                    "dataType": conf_field.get("dataType").upper(),
                    "primaryKey": conf_field.get("primaryKey"),
                    "maxLength": conf_field.get("maxLength"),
                    "minLength": conf_field.get("minLength"),
                    "defaultValue": conf_field.get("defaultValue"),
                    "nullable": conf_field.get("nullable"),
                    "multiple": False
                }
                param_def["metaModel"]["fields"].append(meta_model_field)

            param_defs.append(param_def)

    return param_defs

def build_data_algorithm(conf_algorithm:dict) -> dict:
    data_algorithm = {
        "site": conf_algorithm.get("site"),
        "code": conf_algorithm.get("code"),
        "name": conf_algorithm.get("name"),
        "description": conf_algorithm.get("description"),
        "actorName": conf_algorithm.get("actorName"),
        "handler": conf_algorithm.get("handler"),
        "author": conf_algorithm.get("author"),
        "version": conf_algorithm.get("version"),
        "categories": conf_algorithm.get("categories"),
        "shardable": conf_algorithm.get("shardable"),
        "inputParamDefs": list(),
        "paramDefs": list(),
        "enable": True,
        "createTime": time.strftime("%Y-%m-%d %H:%M:%S")
    }

    #为输入参数赋值
    if "inputParamDefs" in conf_algorithm and len(conf_algorithm["inputParamDefs"]) > 0:
        for conf_input_param_def in conf_algorithm["inputParamDefs"]:
            input_param_def = {
                "code" : conf_input_param_def.get("code"),
                "dataType": conf_input_param_def.get("dataType").upper(),
                "description": conf_input_param_def.get("description"),
                "required": conf_input_param_def.get("required")
            }
            data_algorithm["inputParamDefs"].append(input_param_def)

    #出参入参集合
    if "inputDefs" in conf_algorithm and len(conf_algorithm["inputDefs"]) > 0:
        data_algorithm["paramDefs"].extend(build_param_def(conf_algorithm["inputDefs"], True))

    if "outputDefs" in conf_algorithm and len(conf_algorithm["outputDefs"]) > 0:
        data_algorithm["paramDefs"].extend(build_param_def(conf_algorithm["outputDefs"], False))

    return data_algorithm

def register_algorithm(conf_algorithm:dict, task_handler_names:dict):
    """
    注册算法组件
    :param conf_algorithm:
    :param taskHandlerNames:
    :return:
    """
    # 数据校验
    if "code" not in conf_algorithm or conf_algorithm["code"] == "":
        raise IllegalArgumentException(f"自注册算法组件失败,code为空")

    if "name" not in conf_algorithm or conf_algorithm["name"] == "":
        raise IllegalArgumentException(f"自注册算法组件失败,name为空")

    if "actorName" not in conf_algorithm or conf_algorithm["actorName"] == "":
        raise IllegalArgumentException(f"自注册算法组件失败,actorName为空")

    if "categories" not in conf_algorithm or conf_algorithm["categories"] == "":
        raise IllegalArgumentException(f"自注册算法组件失败,categories为空")

    if "handler" not in conf_algorithm or conf_algorithm["handler"] == "":
        raise IllegalArgumentException(f"自注册算法组件失败,handler为空")
    if conf_algorithm["handler"] not in task_handler_names.keys():
        raise IllegalArgumentException(f"自注册算法组件失败,{conf_algorithm['handler']} 不存在")

    try:
        actor_name = conf_algorithm["actorName"]
        conf_algorithm["version"] = int(actor_name[actor_name.rindex('.v') + 2:])
    except Exception:
        raise IllegalArgumentException(f"自注册算法组件失败,actorName不满足规范,无法获取版本号")
    if conf_algorithm["version"] < 1:
        raise IllegalArgumentException(f"自注册算法组件失败,actorName不满足规范,版本小于1")

    #构建信息
    data_algorithm = build_data_algorithm(conf_algorithm)

    #增加/更新
    if DataAlgorithmClient.exists(data_algorithm["code"], data_algorithm["actorName"]):
        log.info(f"算法组件已存在，不做处理")
        # 暂时没有更新
        # exists_algorithm = DataAlgorithmClient.get(conf_algorithm["code"])
        # if data_algorithm["version"] > exists_algorithm["version"]:
        #     data_algorithm["id"] = exists_algorithm["id"]
        #     data_algorithm["paramDefIds"] = exists_algorithm["paramDefIds"]
        #     DataAlgorithmClient.update(data_algorithm)
        #     log.info(f"更新算法组件")
        # else:
        #     log.warning(f"更新的版本号小于或等于-原始数值:{exists_algorithm['version']}，不做算法组件的处理")
    else:
        DataAlgorithmClient.add(data_algorithm)
        log.info(f"自动注册算法组件")

def registry_after_started():
    """
    服务启动后,检测是否需要注册服务
    :return:
    """

    from registry import ALGORITHM_PROPERTIES
    conf_algorithm = ALGORITHM_PROPERTIES

    if conf_algorithm is None or len(conf_algorithm) == 0:
        # 说明没有yml文件,不进行注册处理
        log.info("没有配置algorithm.yml,不进行算法组件自动注册")
        return

    if not check_ficus_online():
        log.warning("启动时没有可用的ficus服务,忽略自动注册")
        return

    # 找到所有的 TaskHandler
    from api.annotation.annotation import TASK_HANDLERS
    if len(TASK_HANDLERS) == 0:
        log.info("无可用的算法组件,忽略自动注册")
        return

    # 找到所有的taskHandler的名称,key是名字,value是否是messageHandler
    task_handler_names = dict()
    from api.handler.message.IMessageHandler import IMessageHandler
    for name, handler in TASK_HANDLERS.items():
        task_handler_names[name] = isinstance(handler, IMessageHandler)
    register_algorithm(conf_algorithm, task_handler_names)



class AlgoOutWrapper:
    t = None            #这两个对象名字是为了兼容java的ficus那边
    subPath: str = None

    def __init__(self, content, sub_path: str):
        self.t = content
        self.subPath = sub_path

    def content(self):
        return self.t

    def sub_path(self):
        return self.subPath


def WITH(content, sub_path: str):
    """
    封装输出操作，一般是放在OutPutPipe里面使用，如output_stream.output_for_insert(AlgoOutWrapper.WITH('need_insert_data', 'sub_path'))
    :param content: 被操作的内容
    :param sub_path: 子文件，即往outputFd所定义的文件夹下面的子文件里面写数据
    :return: AlgoOutWrapper的dict对象，是为了兼容java的ficus，因为那边接收的是json
    """
    wrapper = AlgoOutWrapper(content, sub_path)
    return wrapper.__dict__



import atexit
import logging
import random
from time import sleep
from urllib.error import URLError, HTTPError

import requests
import consul
from munch import Munch

from config.annotation import Value
from libs.utils import Singleton
from py_eureka_client import eureka_client

from api.exceptions import ServiceNoInstanceException, IllegalArgumentException

_logger = logging.getLogger('FICUS')


class DiscoveryServiceProxy(object):

    def registry(self, server, app_name, instance_host, instance_port, status_page_url, health_check_url,
                 renewal_interval_in_secs, duration_in_secs):
        """
        注册服务
        :return:
        """
        pass

    def registry_discovery(self, server, renewal_interval_in_secs):
        """
        仅仅注册监听器
        :param server:
        :param renewal_interval_in_secs:
        :return:
        """
        pass

    def deregistry(self):
        """
        反注册服务
        :return:
        """
        pass

    def heartbeat(self):
        """
        心跳
        :return:
        """
        pass

    def check_instance_available(self, app_name):
        """
        判断某个服务是否有实例
        :param app_name:
        :return:
        """
        pass

    def do_service(self, service="", return_type="json", app_name="sobeyficus",
                   prefer_ip=False, prefer_https=False,
                   method="GET", headers=None, params=None,
                   data=None, timeout=None, auth=True):
        """
        调用其他服务
        :param service:
        :param return_type:
        :param app_name:
        :param prefer_ip:
        :param prefer_https:
        :param method:
        :param headers:
        :param params:
        :param data:
        :param timeout:
        :param auth:
        :return:
        """
        pass


@Value("${discovery.type:eureka}")
def discovery_type(self):
    pass


class DiscoveryServiceProxyFactory(Singleton):
    __instance = None

    def get_discovery_service_proxy(self, mode: str = None) -> DiscoveryServiceProxy:
        if self.__instance:
            return self.__instance

        if mode is None:
            mode = discovery_type()

        if mode == "eureka":
            self.__instance = EurekaDiscoveryServiceProxy.instance()
        else:
            self.__instance = ConsulDiscoveryServiceProxy.instance()

        return self.__instance


class EurekaDiscoveryServiceProxy(DiscoveryServiceProxy, Singleton):
    def registry(self, server, app_name, instance_host, instance_port, status_page_url, health_check_url,
                 renewal_interval_in_secs, duration_in_secs):
        eureka_client.init(eureka_server=server,
                           app_name=app_name,
                           # 当前组件的主机名，可选参数，如果不填写会自动计算一个，如果服务和 eureka 服务器部署在同一台机器，请必须填写，否则会计算出 127.0.0.1
                           instance_host=instance_host,
                           instance_port=instance_port,
                           # 调用其他服务时的高可用策略，可选，默认为随机
                           ha_strategy=eureka_client.HA_STRATEGY_STICK,
                           status_page_url=status_page_url,
                           health_check_url=health_check_url,
                           renewal_interval_in_secs=renewal_interval_in_secs,
                           duration_in_secs=duration_in_secs)

    def registry_discovery(self, server, renewal_interval_in_secs):
        eureka_client.init_discovery_client(eureka_server=server, renewal_interval_in_secs=renewal_interval_in_secs)

    def deregistry(self):
        # 这个不需要
        pass

    def check_instance_available(self, app_name):

        from config import work_on_k8s
        if not work_on_k8s:
            # 如果没有在k8s上,就去从注册中心上找
            cli = eureka_client.get_discovery_client()
            if cli is None:
                raise Exception("Discovery Client has not initialized. ")
            app = cli.applications.get_application(app_name.upper())
            if app.instances is None or len(app.instances) == 0:
                raise ServiceNoInstanceException(f"{app_name}服务没有找到可用的实例")
        # 如果在k8s上,就不需要去判断

    def heartbeat(self):
        # 这个不需要
        pass

    def do_service(self, service="", return_type="json", app_name="sobeyficus", prefer_ip=False, prefer_https=False,
                   method="GET", headers=None, params=None, data=None, timeout=None, auth=True):
        def walk_using_requests(url):
            from client.ClientAuth import ClientAuth
            global r
            try:
                if method.lower() == "get":
                    r = requests.get(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                     timeout=timeout, params=params)
                elif method.lower() == "post":
                    r = requests.post(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                      timeout=timeout, params=params)
                elif method.lower() == "put":
                    r = requests.put(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                     timeout=timeout, params=params)
                elif method.lower() == "patch":
                    r = requests.patch(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                       timeout=timeout, params=params)
                elif method.lower() == "delete":
                    r = requests.delete(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                        timeout=timeout, params=params)
                else:
                    raise IllegalArgumentException(f"不支持的Rest操作:{method}")
                r.raise_for_status()
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
                from urllib.error import HTTPError
                raise HTTPError(url, 500, "连接错误或超时", str(e), None)

            try:
                if return_type is None or return_type.lower() == "none":
                    return None

                if len(r.content) <= 0:
                    return None

                if return_type.lower() in ("json", "dict", "dictionary"):
                    r.encoding = 'utf-8'
                    return r.json()
                else:
                    r.encoding = 'utf-8'
                    return r.text
            finally:
                r.close()

        cli = eureka_client.get_discovery_client()
        if cli is None:
            raise Exception("Discovery Client has not initialized. ")

        return cli.walk_nodes(app_name, service, prefer_ip, prefer_https, walk_using_requests)


class ConsulDiscoveryServiceProxy(DiscoveryServiceProxy, Singleton):
    __cons: consul.Consul = None
    __service_id = None

    def registry(self, server: str, app_name, instance_host, instance_port, status_page_url, health_check_url,
                 renewal_interval_in_secs, duration_in_secs):
        splited = server.split(':')
        self.__cons = consul.Consul(host=splited[0], port=int(splited[1]))
        self.__service_id = f"{app_name}-{instance_port}"
        index = 1
        while True:
            if index > 10:
                _logger.error(f'consul host is down, already retry:{index} times,Exit')
                exit(-1)
            try:
                self.__cons.agent.service.register(name=app_name, address=instance_host,
                                                   port=instance_port, service_id=self.__service_id,
                                                   check=consul.Check.http(
                                                       f"http://{instance_host}:{instance_port}{health_check_url}",
                                                       f'{renewal_interval_in_secs}s',
                                                       '5s',
                                                       f'{duration_in_secs}s'))  # integrated service registration <3
                break
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout, consul.ConsulException):
                _logger.warn(f'consul host ({server}) is down, reconnecting... retry:{index}')
                index = index + 1
                sleep(0.5)

    def registry_discovery(self, server, renewal_interval_in_secs):
        """这个不需要实现"""
        pass

    def check_instance_available(self, app_name):
        node = self.__get_available_service(app_name)
        if node is None:
            raise ServiceNoInstanceException(f"{app_name}服务没有找到可用的实例")

    def deregistry(self):
        _logger.info(f"deregistry serivce: {self.__service_id}")
        if self.__cons is not None:
            self.__cons.agent.service.deregister(self.__service_id)

    def heartbeat(self):
        # 这个不需要
        pass

    def do_service(self, service="", return_type="json", app_name="sobeyficus", prefer_ip=True, prefer_https=False,
                   method="GET", headers=None, params=None, data=None, timeout=None, auth=True):

        def walk_using_requests(url):
            from client.ClientAuth import ClientAuth
            global r
            try:
                if method.lower() == "get":
                    r = requests.get(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                     timeout=timeout, params=params)
                elif method.lower() == "post":
                    r = requests.post(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                      timeout=timeout, params=params)
                elif method.lower() == "put":
                    r = requests.put(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                     timeout=timeout, params=params)
                elif method.lower() == "patch":
                    r = requests.patch(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                       timeout=timeout, params=params)
                elif method.lower() == "delete":
                    r = requests.delete(url, json=data, auth=ClientAuth() if auth else None, headers=headers,
                                        timeout=timeout, params=params)
                else:
                    raise IllegalArgumentException(f"不支持的Rest操作:{method}")
                r.raise_for_status()
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
                from urllib.error import HTTPError
                raise HTTPError(url, 500, "连接错误或超时", str(e), None)

            try:
                if return_type is None or return_type.lower() == "none":
                    return None

                if len(r.content) <= 0:
                    return None

                if return_type.lower() in ("json", "dict", "dictionary"):
                    r.encoding = 'utf-8'
                    return r.json()
                else:
                    r.encoding = 'utf-8'
                    return r.text
            finally:
                r.close()

        # 找到一个可用的服务地址
        node = self.__get_available_service(app_name)
        error_nodes = []
        while node is not None:
            try:
                url = self.__generate_service_url(node, prefer_ip, prefer_https)
                if service.startswith("/"):
                    url = url + service[1:]
                else:
                    url = url + service
                return walk_using_requests(url)
            except (HTTPError, URLError):
                _logger.warn("do service %s in node [%s] error, use next node." % (service, node.ID))
                error_nodes.append(node.ID)
                node = self.__get_available_service(app_name, error_nodes)

        raise URLError("Try all up instances in registry, but all fail")

    # 思路就是去consul上去获取服务的实例列表
    def __get_available_service(self, application_name, ignore_instance_ids=None):
        # TODO 这里其实还应该增加一个简单的本地缓存的,否则请求consul太频繁
        index, instances = self.__cons.health.service(application_name, passing=True)
        up_instances = []

        if ignore_instance_ids is None or len(ignore_instance_ids) == 0:
            up_instances.extend(instances)
        else:
            for ins in instances:
                if ins["Service"]["ID"] not in ignore_instance_ids:
                    up_instances.append(ins)

        if len(up_instances) == 0:
            # no up instances
            return None
        elif len(up_instances) == 1:
            # only one available instance, then doesn't matter which strategy is.
            instance = Munch(up_instances[0]['Service'])
            # self.__ha_cache[application_name] = instance.instanceId
            return instance

        # 随机选一个
        def random_one(instances):
            if len(instances) == 1:
                idx = 0
            else:
                idx = random.randint(0, len(instances) - 1)
            selected_instance = Munch(instances[idx]['Service'])
            return selected_instance

        return random_one(up_instances)

    def __generate_service_url(self, instance, prefer_ip, prefer_https):
        if instance is None:
            return None
        schema = "http"
        port = 0
        if instance.Port and (len(instance.Tags) == 0 or "secure=false" in instance.Tags):
            schema = "http"
            port = instance.Port
        else:
            schema = "https"
            port = instance.Port

        host = instance.Address

        return "%s://%s:%d/" % (schema, host, port)


def discovery_service_proxy() -> DiscoveryServiceProxy:
    return DiscoveryServiceProxyFactory.instance().get_discovery_service_proxy()


@atexit.register
def deregistry():
    discovery_service_proxy().deregistry()



"""
针对FD的一些操作, 现已重构到 FactDatasourceProxyService.py 中
2020-09-16 16:21:26 sun 不得行, 还是需要
"""
import requests
from munch import Munch

from api.exceptions import ServiceInnerException, IllegalArgumentException, AuthException
from client import check_instance_avaliable, do_service


def fd(fd_code):
    """
    获取某一个FD对象
    :param fd_code: fd的唯一code
    :return: FD对象
    """
    check_instance_avaliable()

    try:
        r = do_service(f"/remote/fd-service/{fd_code}")

        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def exists_fds(fds):
    """
    判断fd是否存在
    :param fds:
    :return:
    """
    if fds is None:
        return False

    if not isinstance(fds, list):
        raise IllegalArgumentException("检测fd是否存在失败,输入参数不是一个list")

    check_instance_avaliable()

    try:
        r = do_service(f"/remote/fd-service/exists", method="post", data=fds,return_type = "bool")
        if r is not None:
            return bool(r)
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



import threading


class ReadWriteLock:
    """A lock object that allows many simultaneous "read-locks", but
    only one "write-lock"."""

    def __init__(self):
        self._read_ready = threading.Condition(threading.Lock())
        self._readers = 0

    def acquire_read(self):
        """Acquire a read-lock. Blocks only if some thread has
        acquired write-lock."""
        self._read_ready.acquire()
        try:
            self._readers += 1
        finally:
            self._read_ready.release()

    def release_read(self):
        """Release a read-lock."""
        self._read_ready.acquire()
        try:
            self._readers -= 1
            if not self._readers:
                self._read_ready.notifyAll()
        finally:
            self._read_ready.release()

    def acquire_write(self):
        """Acquire a write lock. Blocks until there are no
        acquired read- or write-locks."""
        self._read_ready.acquire()
        while self._readers > 0:
            self._read_ready.wait()

    def release_write(self):
        """Release a write-lock."""
        self._read_ready.release()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: tangjy
# Created on 2021/02/02
from factdatasource.execptions import NotSupportedFDException
from factdatasource.persistence.jdbc import SqlWrap
from factdatasource.persistence.jdbc.dialect.MysqlJdbcFactDatasourceContext import MysqlJdbcFactDatasourceContext


class VernoxntsJdbcFactDatasourceContext(MysqlJdbcFactDatasourceContext):

    def _delete_all_sql(self):
        raise NotSupportedFDException('VernoxNTS不支持该操作')

    def _delete_sql(self, query: str):
        raise NotSupportedFDException('VernoxNTS不支持该操作')

    def _generate_update_sql(self, table, result: dict) -> SqlWrap:
        raise NotSupportedFDException('VernoxNTS不支持该操作')


server_ip = None
server_port = 5000
config_name = "sobey-cube"
config_profile = "release"
config_server_id = "sobeyficus-config-server"
config_fail_fast = True
eureka_default_zone = "http://sobeyficus:JXTYp9icQaTzs4@sobeyficus-eureka:8765/eureka/"
application_name = None
actor_name = None
spring_profiles_active = "default"

# 是否工作在K8S上, 如果在K8S上,那么在调用外部服务的时候,可以不通过注册中心,直接走K8S的服务发现
work_on_k8s = False
k8s_auth_server_service = "http://sc-auth-service.sobeyficus:80"        # 认证中心的完整地址
k8s_config_server_service = "http://config-server.sobeyficus:8777"        # 配置中心服务的完整地址
k8s_ficus_server_service = "http://sobeyficus.sobeyficus:8040"          # ficus核心服务的完整地址
k8s_schedule_server_service = "http://schedule-server.sobeyficus:8061"      # 调度服务的完整地址

def find_host_ip():
    """
    获取本机的IP地址
    :return:
    """
    import socket
    import config
    from urllib.parse import urlparse

    if config.server_ip is not None:
        return config.server_ip

    global s
    try:
        # 创建一个临时的socket连接
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # 向自己发送一个udp的请求
        s.connect(('8.8.8.8', 80))
        # 送结果中获取到真实的ip地址
        config.server_ip = s.getsockname()[0]
    except Exception:
        pass
    finally:
        s.close()

    # 有时候通过8.8.8.8不连外网获取不到IP，所以这里再通过向eureka发送一个udp包来获取
    if config.server_ip is None and config.eureka_default_zone is not None:
        urls = config.eureka_default_zone.split(',')
        for url in urls:
            try:
                result = urlparse(url)
                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                # 发送一个udp的请求
                s.connect((result.hostname, result.port or 80))
                config.server_ip = s.getsockname()[0]
                return config.server_ip
            except Exception:
                pass
            finally:
                s.close()

    if config.server_ip is None:
        raise RuntimeError('获取本机IP失败')
    return config.server_ip



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from stomp.connect import stomp, StompConnection11

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.dao.FactDatasource import DatasourceListener, MultipleBaseDatasource, BaseDatasource
from factdatasource.execptions import DatasourceNotFoundException, FDExecuteException
from libs.utils import Singleton


class MultipleStompDatasource(DatasourceListener, MultipleBaseDatasource, Singleton):
    """
    管理fd中的所有STOMP类型的数据源  stomp类的数据源在ficus中的类型是JMS
    """

    def __init__(self):
        # source_name --> StompDatasource
        self._target_dataSources = dict()

    def get_datasource_type(self):
        """
        获取数据源类型
        :return:
        """
        return FactDatasourceTypeEnum.JMS

    def add_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        添加一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name or not url:
            raise IllegalArgumentException(f'添加数据源参数错误:source_name、url都不能为空')

        for target in self._target_dataSources.values():
            if target.url == url and target.credentials == credentials:
                # url 和 credentials 都相同，说明是同一个数据库连接就不再重复创建数据源了
                target.add_source_name(source_name)
                self._target_dataSources[source_name] = target
                return

        # 开始创建数据源
        stomp_data_source = StompDatasource(source_name, url, credentials)
        stomp_data_source.start()
        self._target_dataSources[source_name] = stomp_data_source

    def update_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        修改一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name or not url:
            raise IllegalArgumentException(f'修改数据源参数错误:source_name、url都不能为空')

        self.delete_datasource_type(source_name)
        self.add_datasource_type(source_name, url, credentials)

    def delete_datasource_type(self, source_name: str):
        """
        删除一个数据源
        :param source_name:
        :return:
        """
        target: StompDatasource = self._target_dataSources.pop(source_name, None)
        if target:
            if target.only_one_source():
                target.close_client()
            else:
                target.remove_source_name(source_name)

    def get_data_source(self):
        """
        获取数据源的基本信息
        :return:
        """
        source_name = self.determine_current_lookup_key()
        if not source_name:
            raise FDExecuteException('未设置操作源,无法获取数据源信息。')

        if source_name in self._target_dataSources.keys():
            return self._target_dataSources[source_name]
        raise DatasourceNotFoundException('未发现数据源%s。' % source_name)

    def get_client(self):
        """
        获取数据库操作的client
        :return:
        """
        target: StompDatasource = self.get_data_source()
        return target.get_client()

    def close_client(self):
        """
        关闭客户端，关闭连接
        :return:
        """
        target: StompDatasource = self.get_data_source()
        return target.close_client()


class StompDatasource(BaseDatasource):

    def __init__(self, source_name, url, credentials):
        self.init(source_name, url, credentials)
        self.client: StompConnection11 = None

    def start(self):
        """
        :return:
        """
        url = self._parse_url()
        con_param = {'wait': True}
        if self.credentials:
            credentials_list = self.credentials.split(':')
            con_param['username'] = credentials_list[0]
            con_param['passcode'] = credentials_list[1] if len(credentials_list) > 1 else ''
        self.client = stomp.Connection(url)
        self.client.connect(**con_param)

    def get_client(self):
        return self.client

    def close_client(self):
        self.client.disconnect()

    def _parse_url(self) -> list:
        """
        分析url，构造连接stomp的url
        url
        :return: [('localhost', 61613),('localhost', 61613)]
        """
        result = []
        if ',' in self.url:
            # 说明是多个
            for url_port in self.url.split(','):
                url_list = url_port.split(':')
                port = int(url_list[1]) if len(url_list) > 1 else 61613
                result.append((str(url_list[0]), port))
        else:
            url_list = self.url.split(':')
            port = int(url_list[1]) if len(url_list) > 1 else 61613
            result.append((str(url_list[0]), port))
        return result



import requests
from munch import Munch
from api.exceptions import ServiceInnerException, IllegalArgumentException, AuthException
from client import check_instance_avaliable, do_service


def get(site, project_code, code):
    """
    获取事实库的定义
    :param site:
    :param project_code:
    :param code:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"/remote/fd-service-manage/{site}/{project_code}/{code}")

        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



import logging

from celery.signals import task_revoked

from api.model.ResultVO import ResultVO, FAIL_CODE
from cloudcelery import celery
from schedule import TriggerActor
from munch import Munch
import json

log = logging.getLogger('Ficus')


@celery.task(name='tasks.on_request', bind=True, max_retries=2, default_retry_delay=1 * 6)
def on_request(self, protocol):
    """
    从celery接收协议
    :param self:
    :param protocol:
    :return:
    """
    log.info(f"从celery中获取到任务:{protocol}")
    try:
        # 子进程中也开启eureka_client的刷新,因为celery也会起子进程,所以只能在这里初始化了
        from discovery import discovery_service_proxy
        import config
        discovery_service_proxy().registry_discovery(config.eureka_default_zone, renewal_interval_in_secs=4)
    except:
        pass

    try:
        # 在celery环境中,需要在celery的子进程中也初始化这个监听
        # 这里要提前获取一次,便于提前加载和开启数据库连接以及FD的变化监听
        from factdatasource import FactDatasourceProxyService
        FactDatasourceProxyService.fd_client_proxy()
        log.info("完成FD服务监听加载")
    except:
        pass

    body = Munch(json.loads(protocol))

    result: ResultVO = TriggerActor.handle_trigger(body, True)  # ResultVO
    log.info(f"任务执行完成，jobId:{body.jobId} logId:{body.logId}")

    if result.code == FAIL_CODE:
        # 让celery的任务也置失败
        raise RuntimeError(result.msg)

    return {'status': True, 'data': result.to_dict()}


@task_revoked.connect(sender=on_request,dispatch_uid="cloudcelery.CeleryOnRequest.on_request")
def on_revoke(request=None, terminated=None, signum=None, expired=None, **kwargs):

    if str(signum) != 'Signals.SIGKILL':
        return

    log.info(f"celery服务器:{request.hostname} 接收到任务:{request.id} 的取消事件(terminated:{terminated},expired:{expired})")
    log_id = request.reply_to

    try:
        from handlers import revoke_handler
    except Exception as e:
        # 说明可能没有这个方法
        return

    try:
        # 子进程中也开启eureka_client的刷新,因为celery也会起子进程,所以只能在这里初始化了
        from discovery import discovery_service_proxy
        import config
        discovery_service_proxy().registry_discovery(config.eureka_default_zone, renewal_interval_in_secs=4)
    except:
        pass

    # 这里就根据task_id去查询 log_id,job_id, task_handler
    from client import ScheduleJobTaskLogClient
    from api.handler.ICacheAbleHandler import CacheAbleHandlerHolder
    task_log = ScheduleJobTaskLogClient.get_task_log_by_id(log_id)

    # 这说明有这个方法了
    try:
        actorParam = Munch(json.loads(task_log.actorParam))
        #设置缓存需要的key
        CacheAbleHandlerHolder.get_handler().set_local_code(actorParam["site_"] + "_" + actorParam["projectCode_"] + "_" + actorParam["code_"])
        CacheAbleHandlerHolder.get_handler().set_process_id(actorParam.get("__processLogId__"))

        revoke_handler(task_log.actorCode, task_log.actorHandler, actorParam["code_"],
                       actorParam["projectCode_"], actorParam["site_"], task_log.id, task_log.jobId,
                       task_log.messageId, expired)
    except Exception as e:
        log.warning(f"任务:{task_log},取消回调失败:{str(e)}")



#!/usr/bin/env python
from client.ClientAuth import ClientAuth
from discovery import discovery_service_proxy

FICUS_APP_NAME = "sobeyficus"


def do_service(service="", return_type="json", app_name=FICUS_APP_NAME,
               prefer_ip=False, prefer_https=False,
               method="GET", headers=None, params=None,
               data=None, timeout=None, auth=True, deep=0):
    try:
        return discovery_service_proxy().do_service(service, return_type, app_name, prefer_ip, prefer_https, method,
                                                headers, params, data, timeout, auth)
    except requests.exceptions.HTTPError as e:
        if e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证的问题
            if auth:
                if deep >= 2:
                    raise e
                else:
                    # 那么这里就重新发一次
                    from client.ClientAuth import cancel_oauth_token
                    cancel_oauth_token()
                    return do_service(service, return_type, app_name, prefer_ip, prefer_https, method, headers, params, data, timeout, auth, deep+1)
            else:
                # 说明是不认证的
                raise e
        else:
            # 其他错误
            raise e


def check_instance_avaliable(app_name=FICUS_APP_NAME):
    discovery_service_proxy().check_instance_available(app_name)


from .ComputeExecutionClient import *
from .DataAlgorithmClient import *
from .DataCrawlClient import *
from .FactDatasourceClient import *
from .HandlerLogClient import *
from .JobScheduleClient import *
from .ScheduleCacheClient import *
from .ScheduleJobTaskLogClient import *
from .FactDatasourceManageClient import *



from api.model.FdInputWrapper import FdInputWrapper

from libs.utils import get_algo_abs_path
from api.model.FactDatasource import FactDatasource
import os


class AlgorithmFdInputWrapper(FdInputWrapper):
    """
    算法fd输入的一个包装
    """
    sub_path: str = None
    input_stream = None


    def __init__(self, fd_code: str, sub_path):
        super().__init__(fd_code)
        self.sub_path = sub_path

    def path(self):
        fd = self.info()
        if fd is None or not isinstance(fd, FactDatasource):
            return None
        if fd.target:
            # 有target说明是文件不可能有subPath
            return get_algo_abs_path(fd.site, fd.connection, fd.target)
        else:
            if self.sub_path:
                return get_algo_abs_path(fd.site, fd.connection, self.sub_path)
            else:
                return get_algo_abs_path(fd.site, fd.connection, '')

    def exists(self):
        path = self.path()
        if path is None:
            return False
        return os.path.exists(path)

    def open(self):
        # 防止改fd改了路径，每次都新开InputStream
        self.close()
        path = self.path()
        if path is None or not os.path.isfile(path):
            return None
        self.input_stream = open(path, 'rb')
        return self.input_stream

    def close(self):
        if self.input_stream is not None:
            self.input_stream.close()
    """
    获取子文件
    """
    def child(self, sub_path: str):
        fd = self.info()
        if not isinstance(fd, FactDatasource) or fd.target:
            # 说明本身就是文件类型的，没有子文件
            return None
        if self.sub_path:
            absolute_path = get_algo_abs_path(fd.site, fd.connection, self.sub_path)
            if not os.path.isdir(absolute_path):
                # 不是文件夹就没有子文件了
                return None
            else:
                directory = self.sub_path
                if not directory.startswith(os.sep):
                    directory = os.sep + directory
                if not directory.endswith(os.sep):
                    directory = directory + os.sep
                if sub_path.startswith(os.sep):
                    sub_path = sub_path[1:]
                # self.sub_path = directory + sub_path
                #return self
                return AlgorithmFdInputWrapper(self._fact_datasource_code, directory + sub_path)
        else:
            #self.sub_path = sub_path
            #return self
            return AlgorithmFdInputWrapper(self._fact_datasource_code, sub_path)

    """
    返回子文件/文件夹，只返回一层
    """
    def list(self):
        results = []
        abs_path = self.path()
        if abs_path is None:
            return results
        if not os.path.isdir(abs_path):
            # 不是文件夹就没有子文件了
            return results
        else:
            files = os.listdir(abs_path)
            for file in files:
                if file.endswith(os.sep):
                    file = file[:-1]
                directory = ''
                if self.sub_path:
                    directory = self.sub_path
                    if not directory.startswith(os.sep):
                        directory = os.sep + directory
                    if not directory.endswith(os.sep):
                        directory = directory + os.sep
                #self.sub_path = directory + file
                #results.append(self)
                results.append(AlgorithmFdInputWrapper(self._fact_datasource_code, directory + file))
            return results
